/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "standardshader.h"

void _standardShaderUpdateUniforms(standardshader_t *stds) {
  shaderprogram_t *sp;
  ASSERT_NOT_NULL(stds);

  sp = STANDARD_SHADER_SHADER(stds);
  stds->uniformView = shaderGetUniformByName(
    sp, STANDARD_SHADER_UNIFORM_NAME_VIEW
  );
  stds->uniformProjection = shaderGetUniformByName(
    sp, STANDARD_SHADER_UNIFORM_NAME_PROJ
  );
  stds->uniformTexture = shaderGetUniformByName(
    sp, STANDARD_SHADER_UNIFORM_NAME_TEXT
  );
  stds->uniformColor = shaderGetUniformByName(
    sp, STANDARD_SHADER_UNIFORM_NAME_COLOR
  );
  stds->uniformPosition = shaderGetUniformByName(
    sp, STANDARD_SHADER_UNIFORM_NAME_MODEL
  );
}

bool _standardShaderOnAssetItemLoaded(
  void *u, event_t e, const void *s[], const int32_t c
) {
  standardshader_t *ss = (standardshader_t *)u;

  ASSERT_NOT_NULL(ss);
  ASSERT_EQUAL(e, ASSET_MANAGER_EVENT_ITEM_LOADED);
  ASSERT_NOT_NULL(s);
  ASSERT_EQUAL(c, 2);

  if(s[1] != ss->shaderProgram) return true;

  eventManagerUnsubscribe(&((assetmanager_t *)s[0])->events, ss->onItemLoaded);
  _standardShaderUpdateUniforms(ss);
  return true;
}

void standardShaderInit(standardshader_t *stds, assetmanager_t *assetManager) {
  ASSERT_NOT_NULL(stds);
  ASSERT_NOT_NULL(assetManager);

  // Begin loading the shader.
  stds->holder = assetManagerHolderCreate(assetManager);
  stds->shaderProgram = assetManagerLoadShader(
    assetManager, stds->holder,
    STANDARD_SHARER_FILE_VERTEX, STANDARD_SHADER_FILE_FRAGMENT
  );

  // Update uniforms
  if(assetManagerItemIsFinished(stds->shaderProgram)) {
    _standardShaderUpdateUniforms(stds);
  } else {
    stds->onItemLoaded = eventManagerSubscribe(
      &assetManager->events, ASSET_MANAGER_EVENT_ITEM_LOADED,
      stds, &_standardShaderOnAssetItemLoaded
    );
  }
}

void standardShaderUse(standardshader_t *stds) {
  ASSERT_NOT_NULL(stds);
  if(!assetManagerItemIsFinished(stds->shaderProgram)) return;
  shaderUse(STANDARD_SHADER_SHADER(stds));
}

void standardShaderSetCamera(standardshader_t *stds, camera_t *camera) {
  ASSERT_NOT_NULL(stds);
  ASSERT_NOT_NULL(camera);
  if(!assetManagerItemIsFinished(stds->shaderProgram)) return;

  shaderUseCamera(STANDARD_SHADER_SHADER(stds),
    stds->uniformView, stds->uniformProjection, camera
  );
}

void standardShaderSetTexture(standardshader_t *stds, texture_t *texture) {
  ASSERT_NOT_NULL(stds);
  ASSERT_NOT_NULL(texture);
  if(!assetManagerItemIsFinished(stds->shaderProgram)) return;

  textureBind(texture, 0);
  shaderUseTexture(STANDARD_SHADER_SHADER(stds), stds->uniformTexture, 0);
}

void standardShaderSetColor(standardshader_t *stds, pixel_t color) {
  ASSERT_NOT_NULL(stds);
  if(!assetManagerItemIsFinished(stds->shaderProgram)) return;

  shaderUseColor(STANDARD_SHADER_SHADER(stds), stds->uniformColor, color);
}

void standardShaderSetPosition(standardshader_t *stds, matrix_t *matrix) {
  ASSERT_NOT_NULL(stds);
  ASSERT_NOT_NULL(matrix);
  if(!assetManagerItemIsFinished(stds->shaderProgram)) return;

  shaderUseMatrix(STANDARD_SHADER_SHADER(stds), stds->uniformPosition, matrix);
}

void standardShaderDispose(standardshader_t *stds, assetmanager_t *assMan) {
  ASSERT_NOT_NULL(stds);
  ASSERT_NOT_NULL(assMan);

  assetManagerHolderRelease(assMan, stds->holder);
}